# r: timber_design>=0.1.0
# venv: td_migration
import Grasshopper
from compas_timber.connections import JointTopology
from compas_timber.connections import LMiterJoint
from compas_timber.connections import TButtJoint
from compas_timber.connections import XLapJoint

from timber_design.workflow import TopologyRule


class DefaultJointRule(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(self):
        topoRules = []
        topoRules.append(TopologyRule(JointTopology.TOPO_L, LMiterJoint))
        topoRules.append(TopologyRule(JointTopology.TOPO_T, TButtJoint))
        topoRules.append(TopologyRule(JointTopology.TOPO_X, XLapJoint))

        return topoRules



# r: timber_design>=0.1.0
# venv: td_migration
"""Extracts main geometric characteristics of a Beam."""

# flake8: noqa
import Grasshopper
import System
from System.Drawing import Color

from compas.geometry import Line
from compas_rhino.conversions import frame_to_rhino_plane
from compas_rhino.conversions import line_to_rhino
from compas_rhino.conversions import point_to_rhino
from compas_rhino.conversions import box_to_rhino

from timber_design.ghpython.ghcomponent_helpers import list_input_valid_cpython


class BeamDecompose(Grasshopper.Kernel.GH_ScriptInstance):
    RED = Color.FromArgb(255, 255, 100, 100)
    GREEN = Color.FromArgb(200, 50, 220, 100)
    BLUE = Color.FromArgb(200, 50, 150, 255)
    WHITE = Color.FromArgb(255, 255, 255, 255)
    YELLOW = Color.FromArgb(255, 255, 255, 0)
    SCREEN_SIZE = 10
    RELATIVE_SIZE = 0

    def RunScript(self, beam: System.Collections.Generic.List[object], show_frame: bool, show_faces: bool):
        self.show_faces = show_faces if show_faces is not None else False
        self.show_frame = show_frame if show_frame is not None else False
        self.frames = []
        self.rhino_frames = []
        self.scales = []
        self.faces = []
        self.width = []
        self.height = []
        self.centerline = []
        self.shapes = []
        if not list_input_valid_cpython(ghenv, beam, "beam"):
            return
        else:
            for b in beam:
                self.frames.append(b.frame)
                self.rhino_frames.append(frame_to_rhino_plane(b.frame))
                self.scales.append(b.width + b.height)
                self.centerline.append(line_to_rhino(b.centerline))
                self.shapes.append(box_to_rhino(b.shape))
                self.width.append(b.width)
                self.height.append(b.height)
                self.faces.append(b.faces)

            return self.rhino_frames, self.centerline, self.shapes, self.width, self.height

    def DrawViewportWires(self, arg):
        if ghenv.Component.Locked:
            return

        for f, s, faces in zip(self.frames, self.scales, self.faces):
            if self.show_frame:
                self._draw_frame(arg.Display, f, s)
            if self.show_faces:
                self._draw_faces(arg.Display, faces, s)

    def _draw_frame(self, display, frame, scale):
        x = Line.from_point_and_vector(frame.point, frame.xaxis * scale)
        y = Line.from_point_and_vector(frame.point, frame.yaxis * scale)
        z = Line.from_point_and_vector(frame.point, frame.zaxis * scale)
        display.DrawArrow(line_to_rhino(x), self.RED, self.SCREEN_SIZE, self.RELATIVE_SIZE)
        display.DrawArrow(line_to_rhino(y), self.GREEN, self.SCREEN_SIZE, self.RELATIVE_SIZE)
        display.DrawArrow(line_to_rhino(z), self.BLUE, self.SCREEN_SIZE, self.RELATIVE_SIZE)

        x_loc = x.end + x.vector * scale * 1.1
        y_loc = y.end + y.vector * scale * 1.1
        z_loc = z.end + z.vector * scale * 1.1
        display.Draw2dText("X", self.RED, point_to_rhino(x_loc), True, 16, "Verdana")
        display.Draw2dText("Y", self.GREEN, point_to_rhino(y_loc), True, 16, "Verdana")
        display.Draw2dText("Z", self.BLUE, point_to_rhino(z_loc), True, 16, "Verdana")

    def _draw_faces(self, display, faces, scale):
        for index, face in enumerate(faces):
            normal = Line.from_point_and_vector(face.point, face.normal * scale)
            text = str(index)
            display.Draw2dText(text, self.WHITE, point_to_rhino(face.point), True, 16, "Verdana")
            display.DrawArrow(line_to_rhino(normal), self.YELLOW, self.SCREEN_SIZE, self.RELATIVE_SIZE)



# r: timber_design>=0.1.0
# venv: td_migration
# flake8: noqa
import Grasshopper
import System

from compas_rhino.conversions import point_to_rhino
from timber_design.ghpython.ghcomponent_helpers import list_input_valid_cpython


class ShowElementIndex(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(self, model):
        self.pt = []
        self.txt = []

        if not list_input_valid_cpython(ghenv, model, "Model"):
            return
        for element in model.beams + model.plates:
            self.pt.append(point_to_rhino(element.midpoint))
            self.txt.append(str(element.key))

    def DrawViewportWires(self, arg):
        if ghenv.Component.Locked:
            return
        col = System.Drawing.Color.FromArgb(255, 255, 255, 255)
        # https://developer.rhino3d.com/api/RhinoCommon/html/M_Rhino_Display_DisplayPipeline_Draw2dText_5.htm
        for p, t in zip(self.pt, self.txt):
            arg.Display.Draw2dText(t, col, p, True, 16, "Verdana")



import pytest
from compas.geometry import Line
from compas.geometry import Point
from compas.geometry import Polyline
from compas.geometry import Vector
from compas.geometry import Frame

from timber_design.workflow import JointRuleSolver
from timber_design.workflow import get_clusters_from_model
from compas_timber.connections import JointTopology
from compas_timber.connections import LButtJoint
from compas_timber.connections import LLapJoint
from compas_timber.connections import LFrenchRidgeLapJoint
from compas_timber.connections import TButtJoint
from compas_timber.connections import XLapJoint
from compas_timber.connections import LMiterJoint
from compas_timber.connections import PlateTButtJoint
from compas_timber.connections import PlateLButtJoint
from compas_timber.connections import PlateMiterJoint
from compas_timber.connections import JointCandidate
from compas_timber.connections import PlateJointCandidate
from compas_timber.analyzers import Cluster
from compas_timber.elements import Beam
from compas_timber.elements import Plate
from timber_design.workflow import DirectRule
from timber_design.workflow import CategoryRule
from timber_design.workflow import TopologyRule
from compas_timber.model import TimberModel


@pytest.fixture
def beams():
    """
    0 <=> 1:L
    1 <=> 2:T
    2 <=> 3:L
    3 <=> 0:X

    """
    w = 0.2
    h = 0.2
    lines = [
        Line(Point(-1, 0, 0), Point(1, 0, 0)),
        Line(Point(1, 0, 0), Point(1, 2, 0)),
        Line(Point(1, 1, 0), Point(0, 1, 0)),
        Line(Point(0, 1, 0), Point(0, -1, 0)),
    ]
    return [Beam.from_centerline(line, w, h) for line in lines]


@pytest.fixture
def separated_beams():
    """
    0 <=> 1:L
    1 <=> 2:T
    2 <=> 3:L
    3 <=> 0:X

    """
    w = 0.2
    h = 0.2
    lines = [
        Line(Point(-1, 0, 0), Point(1, 0, 0)),
        Line(Point(1, 0, 0.1), Point(1, 2, 0.1)),
        Line(Point(1, 1, 0), Point(0, 1, 0)),
        Line(Point(0, 1, 0.1), Point(0, -1, 0.1)),
    ]
    return [Beam.from_centerline(line, w, h) for line in lines]


@pytest.fixture
def L_beams():
    """
    0 <=> 1:L
    1 <=> 2:L
    2 <=> 3:L
    3 <=> 0:X

    """
    w = 0.2
    h = 0.2
    lines = [
        Line(Point(-1, 0, 0), Point(1, 0, 0)),
        Line(Point(1, 0, 0), Point(1, 1, 0)),
        Line(Point(1, 1, 0), Point(0, 1, 0)),
        Line(Point(0, 1, 0), Point(0, -1, 0)),
    ]
    return [Beam.from_centerline(line, w, h) for line in lines]


@pytest.fixture
def L_beams_separated():
    """
    0 <=> 1:L
    1 <=> 2:L
    2 <=> 3:L
    3 <=> 0:X

    """
    w = 0.2
    h = 0.2
    lines = [
        Line(Point(-1, 0, 0), Point(1, 0, 0)),
        Line(Point(1, 0, 0.1), Point(1, 1, 0.1)),
        Line(Point(1, 1, 0), Point(0, 1, 0)),
        Line(Point(0, 1, 0.1), Point(0, -1, 0.1)),
    ]
    return [Beam.from_centerline(line, w, h) for line in lines]


@pytest.fixture
def Y_beams():
    """
    0 <=> 1:L
    1 <=> 2:L
    2 <=> 3:L

    """
    w = 0.2
    h = 0.2
    lines = [
        Line(Point(0, 0, 0), Point(1, 0, 0)),
        Line(Point(0, 0, 0), Point(0, 1, 0)),
        Line(Point(0, 0, 0), Point(-1, -1, 0)),
    ]
    return [Beam.from_centerline(line, w, h) for line in lines]


@pytest.fixture
def K_beams():
    """
    0 <=> 1:Y
    1 <=> 2:Y
    2 <=> 3:Y

    """
    w = 0.2
    h = 0.2
    lines = [
        Line(Point(0, 0, 0), Point(1, 0, 0)),
        Line(Point(0, -1, 0), Point(0, 1, 0)),
        Line(Point(0, 0, 0), Point(-1, -1, 0)),
    ]
    return [Beam.from_centerline(line, w, h) for line in lines]


def test_joints_from_beams_and_rules_with_no_max_distance(separated_beams):
    rules = [
        TopologyRule(JointTopology.TOPO_L, LMiterJoint),
        TopologyRule(JointTopology.TOPO_T, TButtJoint),
        TopologyRule(JointTopology.TOPO_X, XLapJoint),
    ]
    model = TimberModel()
    model.add_elements(separated_beams)
    solver = JointRuleSolver(rules)
    with pytest.raises(ValueError):
        errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len(model.joints) == 0


def test_joints_from_beams_and_rules_with_max_distance_rule(separated_beams):
    rules = [
        TopologyRule(JointTopology.TOPO_L, LMiterJoint),
        TopologyRule(JointTopology.TOPO_T, TButtJoint, max_distance=0.15),
        TopologyRule(JointTopology.TOPO_X, XLapJoint),
    ]
    model = TimberModel()
    model.add_elements(separated_beams)
    solver = JointRuleSolver(rules)
    _, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len([j for j in model.joints if not isinstance(j, JointCandidate)]) == 1
    assert len(unjoined_clusters) == 3


def test_joints_from_beams_and_rules_with_max_distance_model(separated_beams):
    model = TimberModel()
    model.add_elements(separated_beams)
    rules = [
        TopologyRule(JointTopology.TOPO_L, LMiterJoint),
        TopologyRule(JointTopology.TOPO_T, TButtJoint, max_distance=0.05),
        TopologyRule(JointTopology.TOPO_X, XLapJoint),
    ]
    model = TimberModel()
    model.add_elements(separated_beams)
    solver = JointRuleSolver(rules, max_distance=0.15)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len([j for j in model.joints if not isinstance(j, JointCandidate)]) == 3
    assert len(unjoined_clusters) == 1


def test_direct_rule_matches_cluster(beams):
    rule = DirectRule(LMiterJoint, beams[:2])
    cluster_a = Cluster([JointCandidate(*beams[:2])])
    cluster_b = Cluster([JointCandidate(*beams[1:3])])
    assert rule._matches_cluster(cluster_a) is True
    assert rule._matches_cluster(cluster_b) is False


def test_direct_rule_get_joint(beams):
    rules = [
        DirectRule(LMiterJoint, [beams[0], beams[1]]),
        DirectRule(LMiterJoint, [beams[2], beams[3]]),
        DirectRule(TButtJoint, [beams[2], beams[1]]),
        DirectRule(XLapJoint, [beams[3], beams[0]]),
    ]
    model = TimberModel()
    model.add_elements(beams)
    solver = JointRuleSolver(rules)
    _, _ = solver.apply_rules_to_model(model)
    assert set([j.__class__.__name__ for j in model.joints]) == set(["LMiterJoint", "TButtJoint", "XLapJoint"])


def test_joint_type_check_elements_compatibility(beams):
    rules = [DirectRule(LMiterJoint, [beams[0], beams[1]])]
    model = TimberModel()
    model.add_elements(beams)
    solver = JointRuleSolver(rules)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len([j for j in model.joints if not isinstance(j, JointCandidate)]) == 1


def test_joint_type_check_elements_compatibility_bad_normal(beams):
    new_frame = Frame(beams[1].frame.point, beams[1].frame.xaxis, Vector(1, 1, 1))
    new_beam = Beam(frame=new_frame, length=beams[1].length, width=beams[1].width, height=beams[1].height)
    beams[1] = new_beam  # replace beam 1 with a beam with a different frame

    rules = [DirectRule(LFrenchRidgeLapJoint, [beams[0], beams[1]])]

    model = TimberModel()
    model.add_elements(beams)
    solver = JointRuleSolver(rules)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len(errors) == 1
    assert len([j for j in model.joints if not isinstance(j, JointCandidate)]) == 0


def test_joint_type_check_elements_compatibility_bad_dims(beams):
    rules = [DirectRule(LFrenchRidgeLapJoint, [beams[0], beams[1]])]
    beams[1].width = 0.25
    model = TimberModel()
    model.add_elements(beams)
    solver = JointRuleSolver(rules)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len(errors) == 1
    assert len([j for j in model.joints if not isinstance(j, JointCandidate)]) == 0


def test_direct_rule_try_get_joint_max_distance_failed(separated_beams):
    rule = DirectRule(LMiterJoint, [separated_beams[0], separated_beams[1]], max_distance=0.05)
    model = TimberModel()
    model.add_elements(separated_beams)
    solver = JointRuleSolver([rule], max_distance=0.15)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len(errors) == 1
    assert len(list(model.joints)) == 0


def test_direct_rule_try_get_joint_max_distance_success(separated_beams):
    rule = DirectRule(LMiterJoint, [separated_beams[0], separated_beams[1]], max_distance=0.15)
    model = TimberModel()
    model.add_elements(separated_beams)
    solver = JointRuleSolver([rule])
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len(errors) == 0
    assert len(list(model.joints)) == 1


def test_category_rule_try_get_joint(beams):
    for beam in beams:
        beam.attributes["category"] = "A"
    beams[1].attributes["category"] = "B"
    rule = CategoryRule(LMiterJoint, "A", "B")
    model = TimberModel()
    model.add_elements(beams)
    solver = JointRuleSolver([rule])
    errors, unjoined_clusters = solver.apply_rules_to_model(model)

    assert len([j for j in model.joints if not isinstance(j, JointCandidate)]) == 1


def test_category_rule_same_category(beams):
    main = beams[2]
    cross = beams[1]
    for beam in beams:
        beam.attributes["category"] = "A"
    rule = CategoryRule(TButtJoint, "A", "A")
    model = TimberModel()
    model.add_elements([cross, main])
    solver = JointRuleSolver([rule])
    _, _ = solver.apply_rules_to_model(model)
    joint = list(model.joints)[0]
    assert joint.main_beam == main
    assert joint.cross_beam == cross

    model = TimberModel()
    model.add_elements([main, cross])
    solver = JointRuleSolver([rule])
    _, _ = solver.apply_rules_to_model(model)
    joint = list(model.joints)[0]
    assert joint.main_beam == main
    assert joint.cross_beam == cross


def test_topology_rule_try_get_joint(beams):
    rule = TopologyRule(JointTopology.TOPO_L, LMiterJoint)
    model = TimberModel()
    model.add_elements(beams)
    solver = JointRuleSolver([rule])
    _, _ = solver.apply_rules_to_model(model)
    assert len([j for j in model.joints if not isinstance(j, JointCandidate)]) == 2


def test_mixed_rules(L_beams):
    for beam in L_beams:
        beam.attributes["category"] = "A"
    L_beams[1].attributes["category"] = "B"
    rules = [DirectRule(LLapJoint, L_beams[:2]), CategoryRule(LButtJoint, "A", "B"), TopologyRule(JointTopology.TOPO_L, LMiterJoint)]
    model = TimberModel()
    model.add_elements(L_beams)
    solver = JointRuleSolver(rules)
    _, _ = solver.apply_rules_to_model(model)
    assert len(model.joints) == 3
    assert set([joint.__class__.__name__ for joint in model.joints]) == set(["LLapJoint", "LButtJoint", "LMiterJoint"])


def test_different_rules_max_distance(L_beams_separated):
    for beam in L_beams_separated:
        beam.attributes["category"] = "A"
    L_beams_separated[1].attributes["category"] = "B"
    rules = [DirectRule(LLapJoint, L_beams_separated[:2]), CategoryRule(LButtJoint, "A", "B"), TopologyRule(JointTopology.TOPO_L, LMiterJoint)]
    model = TimberModel()
    model.add_elements(L_beams_separated)
    solver = JointRuleSolver(rules)
    with pytest.raises(ValueError):
        errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len([j for j in model.joints if not isinstance(j, JointCandidate)]) == 0


def test_different_rules_max_distance_on_topo_rule(L_beams_separated):
    for beam in L_beams_separated:
        beam.attributes["category"] = "A"
    L_beams_separated[1].attributes["category"] = "B"
    rules = [DirectRule(LLapJoint, L_beams_separated[:2]), CategoryRule(LButtJoint, "A", "B"), TopologyRule(JointTopology.TOPO_L, LMiterJoint, max_distance=0.15)]
    model = TimberModel()
    model.add_elements(L_beams_separated)
    solver = JointRuleSolver(rules)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    # TopologyRule overrides CategoryRule when the latter fails to make a joint, but DirectRule raises error and is not overridden
    assert len([j for j in model.joints if not isinstance(j, JointCandidate)]) == 2
    assert len(unjoined_clusters) == 2
    assert len(errors) == 1  # error because DirectRule fails
    assert set([joint.__class__.__name__ for joint in model.joints]) == set(["LMiterJoint"])  # TopologyRule has generated the joint


def test_different_rules_max_distance_on_category_rule(L_beams_separated):
    for beam in L_beams_separated:
        beam.attributes["category"] = "A"
    L_beams_separated[1].attributes["category"] = "B"
    rules = [DirectRule(LLapJoint, L_beams_separated[:2]), CategoryRule(LButtJoint, "A", "B", max_distance=0.15), TopologyRule(JointTopology.TOPO_L, LMiterJoint)]
    model = TimberModel()
    model.add_elements(L_beams_separated)
    solver = JointRuleSolver(rules)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len(model.joints) == 1
    assert len(unjoined_clusters) == 3
    assert len(errors) == 1  # error because CategoryRule fails
    assert set([joint.__class__.__name__ for joint in model.joints]) == set(["LButtJoint"])  # CategoryRule has generated the joint


def test_different_rules_max_distance_on_direct_rule(L_beams_separated):
    for beam in L_beams_separated:
        beam.attributes["category"] = "A"
    L_beams_separated[1].attributes["category"] = "B"
    rules = [DirectRule(LLapJoint, L_beams_separated[:2], max_distance=0.15), CategoryRule(LButtJoint, "A", "B"), TopologyRule(JointTopology.TOPO_L, LMiterJoint)]
    model = TimberModel()
    model.add_elements(L_beams_separated)
    solver = JointRuleSolver(rules)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len(model.joints) == 1
    assert len(unjoined_clusters) == 3
    assert len(errors) == 0  # NO error because DirectRule succeeds
    assert set([joint.__class__.__name__ for joint in model.joints]) == set(["LLapJoint"])  # DirectRule has generated the joint


def test_different_rules_max_distance_on_rule_solver(L_beams_separated):
    for beam in L_beams_separated:
        beam.attributes["category"] = "A"
    L_beams_separated[1].attributes["category"] = "B"
    rules = [DirectRule(LLapJoint, L_beams_separated[:2]), CategoryRule(LButtJoint, "A", "B"), TopologyRule(JointTopology.TOPO_L, LMiterJoint)]
    model = TimberModel()
    model.add_elements(L_beams_separated)
    solver = JointRuleSolver(rules, max_distance=0.15)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len(model.joints) == 3
    assert len(unjoined_clusters) == 1
    assert len(errors) == 0
    assert set([joint.__class__.__name__ for joint in model.joints]) == set(["LLapJoint", "LButtJoint", "LMiterJoint"])  # All rules have generated the joints


def test_plate_topo_rules():
    polyline_a = Polyline([Point(0, 0, 0), Point(0, 20, 0), Point(10, 20, 0), Point(10, 0, 0), Point(0, 0, 0)])

    plate_a = Plate.from_outline_thickness(polyline_a, 1)

    polyline_b = Polyline([Point(0, 10, 0), Point(10, 10, 0), Point(20, 20, 10), Point(0, 20, 10), Point(0, 10, 0)])

    plate_b = Plate.from_outline_thickness(polyline_b, 1)

    polyline_c = Polyline([Point(10, 0, 0), Point(20, 0, 10), Point(20, 20, 10), Point(10, 10, 0), Point(10, 0, 0)])

    plate_c = Plate.from_outline_thickness(polyline_c, 1)

    rules = [
        TopologyRule(JointTopology.TOPO_EDGE_FACE, PlateTButtJoint),
        TopologyRule(JointTopology.TOPO_EDGE_EDGE, PlateMiterJoint),
    ]
    model = TimberModel()
    model.add_elements([plate_a, plate_b, plate_c])
    solver = JointRuleSolver(rules)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len([j for j in model.joints if not isinstance(j, PlateJointCandidate)]) == 3, "Expected three joints"


def test_plate_category_rules_reverse_topo():
    polyline_a = Polyline([Point(0, 0, 0), Point(0, 20, 0), Point(10, 20, 0), Point(10, 0, 0), Point(0, 0, 0)])
    plate_a = Plate.from_outline_thickness(polyline_a, 1)
    plate_a.attributes["category"] = "A"

    polyline_b = Polyline([Point(0, 10, 0), Point(10, 10, 0), Point(20, 20, 10), Point(0, 20, 10), Point(0, 10, 0)])
    plate_b = Plate.from_outline_thickness(polyline_b, 1)
    plate_b.attributes["category"] = "B"

    polyline_c = Polyline([Point(10, 0, 0), Point(20, 0, 10), Point(20, 20, 10), Point(10, 10, 0), Point(10, 0, 0)])
    plate_c = Plate.from_outline_thickness(polyline_c, 1)
    plate_c.attributes["category"] = "C"

    rules = [
        CategoryRule(PlateTButtJoint, "A", "B"),
        CategoryRule(PlateMiterJoint, "A", "C"),
        CategoryRule(PlateLButtJoint, "B", "C"),
    ]
    model = TimberModel()
    model.add_elements([plate_a, plate_b, plate_c])
    solver = JointRuleSolver(rules)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)

    assert len([j for j in model.joints if not isinstance(j, PlateJointCandidate)]) == 2, "Expected two joints"


def test_plate_category_rules_correct_topo():
    polyline_a = Polyline([Point(0, 0, 0), Point(0, 20, 0), Point(10, 20, 0), Point(10, 0, 0), Point(0, 0, 0)])
    plate_a = Plate.from_outline_thickness(polyline_a, 1)
    plate_a.attributes["category"] = "A"

    polyline_b = Polyline([Point(0, 10, 0), Point(10, 10, 0), Point(20, 20, 10), Point(0, 20, 10), Point(0, 10, 0)])
    plate_b = Plate.from_outline_thickness(polyline_b, 1)
    plate_b.attributes["category"] = "B"

    polyline_c = Polyline([Point(10, 0, 0), Point(20, 0, 10), Point(20, 20, 10), Point(10, 10, 0), Point(10, 0, 0)])
    plate_c = Plate.from_outline_thickness(polyline_c, 1)
    plate_c.attributes["category"] = "C"

    rules = [
        CategoryRule(PlateTButtJoint, "B", "A"),
        CategoryRule(PlateMiterJoint, "A", "C"),
        CategoryRule(PlateLButtJoint, "B", "C"),
    ]
    model = TimberModel()
    model.add_elements([plate_a, plate_b, plate_c])
    solver = JointRuleSolver(rules)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len([j for j in model.joints if not isinstance(j, PlateJointCandidate)]) == 3, "Expected three joints"


def test_plate_rules_priority():
    polyline_a = Polyline([Point(0, 0, 0), Point(0, 20, 0), Point(10, 20, 0), Point(10, 0, 0), Point(0, 0, 0)])
    plate_a = Plate.from_outline_thickness(polyline_a, 1)
    plate_a.attributes["category"] = "A"

    polyline_b = Polyline([Point(0, 10, 0), Point(10, 10, 0), Point(20, 20, 10), Point(0, 20, 10), Point(0, 10, 0)])
    plate_b = Plate.from_outline_thickness(polyline_b, 1)
    plate_b.attributes["category"] = "B"

    polyline_c = Polyline([Point(10, 0, 0), Point(20, 0, 10), Point(20, 20, 10), Point(10, 10, 0), Point(10, 0, 0)])
    plate_c = Plate.from_outline_thickness(polyline_c, 1)
    plate_c.attributes["category"] = "C"

    rules = [
        DirectRule(PlateLButtJoint, [plate_c, plate_b]),
        CategoryRule(PlateTButtJoint, "B", "A"),
        CategoryRule(PlateMiterJoint, "A", "C"),
        CategoryRule(PlateMiterJoint, "B", "C"),
        TopologyRule(JointTopology.TOPO_EDGE_FACE, PlateTButtJoint),
        TopologyRule(JointTopology.TOPO_EDGE_EDGE, PlateMiterJoint),
    ]
    model = TimberModel()
    model.add_elements([plate_a, plate_b, plate_c])
    solver = JointRuleSolver(rules)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len(model.joints) == 3, "Expected three joints"
    assert set([j.__class__.__name__ for j in model.joints]) == set(["PlateLButtJoint", "PlateTButtJoint", "PlateMiterJoint"]), (
        "Expected PlateLButtJoint, PlateTButtJoint, and PlateMiterJoint"
    )

    rules = [
        CategoryRule(PlateTButtJoint, "B", "A"),
        CategoryRule(PlateLButtJoint, "A", "C"),
        CategoryRule(PlateLButtJoint, "B", "C"),
        TopologyRule(JointTopology.TOPO_EDGE_FACE, PlateTButtJoint),
        TopologyRule(JointTopology.TOPO_EDGE_EDGE, PlateMiterJoint),
    ]
    model = TimberModel()
    model.add_elements([plate_a, plate_b, plate_c])
    solver = JointRuleSolver(rules)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len(model.joints) == 3, "Expected three joints"
    assert set([j.__class__.__name__ for j in model.joints]) == set(["PlateLButtJoint", "PlateTButtJoint"]), "Expected PlateLButtJoint, PlateTButtJoint"

    rules = [
        CategoryRule(PlateTButtJoint, "B", "A"),
        CategoryRule(PlateLButtJoint, "A", "C"),
        TopologyRule(JointTopology.TOPO_EDGE_FACE, PlateTButtJoint),
        TopologyRule(JointTopology.TOPO_EDGE_EDGE, PlateMiterJoint),
    ]
    model = TimberModel()
    model.add_elements([plate_a, plate_b, plate_c])
    solver = JointRuleSolver(rules)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len(model.joints) == 3, "Expected three joints"
    assert set([j.__class__.__name__ for j in model.joints]) == set(["PlateLButtJoint", "PlateTButtJoint", "PlateMiterJoint"]), (
        "Expected PlateLButtJoint, PlateTButtJoint, and PlateMiterJoint"
    )


def test_joints_created_with_y_topo_cluster(Y_beams):
    rules = [
        TopologyRule(JointTopology.TOPO_L, LButtJoint),
    ]
    model = TimberModel()
    model.add_elements(Y_beams)

    clusters = get_clusters_from_model(model)
    assert len(clusters) == 1
    assert clusters[0].topology == JointTopology.TOPO_Y
    assert len(clusters[0].joints) == 3

    solver = JointRuleSolver(rules)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len(unjoined_clusters) == 0
    assert len(model.joints) == 3
    assert all([isinstance(j, LButtJoint) for j in model.joints])


def test_joints_created_with_k_topo_cluster(K_beams):
    rules = [
        TopologyRule(JointTopology.TOPO_L, LButtJoint),
        TopologyRule(JointTopology.TOPO_T, TButtJoint),
    ]
    model = TimberModel()
    model.add_elements(K_beams)

    clusters = get_clusters_from_model(model)
    assert len(clusters) == 1
    assert clusters[0].topology == JointTopology.TOPO_K
    assert len(clusters[0].joints) == 3

    solver = JointRuleSolver(rules)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len(unjoined_clusters) == 0
    assert len(model.joints) == 3
    assert any([isinstance(j, LButtJoint) for j in model.joints])
    assert any([isinstance(j, TButtJoint) for j in model.joints])


def test_joints_created_with_k_topo_cluster_l_fails(K_beams):
    rules = [
        TopologyRule(JointTopology.TOPO_T, TButtJoint),
    ]
    model = TimberModel()
    model.add_elements(K_beams)

    clusters = get_clusters_from_model(model)
    assert len(clusters) == 1
    assert clusters[0].topology == JointTopology.TOPO_K
    assert len(clusters[0].joints) == 3

    solver = JointRuleSolver(rules)
    errors, unjoined_clusters = solver.apply_rules_to_model(model)
    assert len(unjoined_clusters) == 1
    assert len(model.joints) == 2
    assert all([isinstance(j, TButtJoint) for j in model.joints])



# r: timber_design>=0.1.0
# venv: td_migration
# flake8: noqa
import Grasshopper
import Rhino

from compas_timber.fabrication import BTLxWriter
from timber_design.ghpython.ghcomponent_helpers import item_input_valid_cpython


class WriteBTLx(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(self, model, path, write: bool):
        if not item_input_valid_cpython(ghenv, model, "Model"):
            return

        btlx = BTLxWriter(file_name=str(Rhino.RhinoDoc.ActiveDoc.Name))

        if write:
            if not item_input_valid_cpython(ghenv, path, "Path"):
                return
            XML = btlx.write(model, path)
        else:
            XML = btlx.model_to_xml(model)

        return XML



# r: timber_design>=0.1.0
# venv: td_migration
# flake8: noqa
import Grasshopper

from compas.scene import Scene

from timber_design.ghpython.ghcomponent_helpers import item_input_valid_cpython


class ShowElementsByType(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(self, model):
        if not item_input_valid_cpython(ghenv, model, "Model"):
            return
        beam_scene = Scene()
        plate_scene = Scene()
        fastener_scene = Scene()

        for beam in model.beams:
            beam_scene.add(beam._geometry or beam.blank)
        for plate in model.plates:
            plate_scene.add(plate._geometry or plate.blank)
        for fastener in model.fasteners:
            fastener_scene.add(fastener.geometry)

        return beam_scene.draw(), plate_scene.draw(), fastener_scene.draw()



# r: timber_design>=0.1.0
# venv: td_migration
"""Defines which Joint type will be applied in the Automatic Joints component for connecting Beams with the given Category attributes. This overrides Topological Joint rules and is overriden by Direct joint rules"""

# flake8: noqa
import inspect
from collections import OrderedDict

import Grasshopper
from System.Windows.Forms import ToolStripSeparator

from compas_timber.connections import PlateJoint
from timber_design.workflow import CategoryRule
from timber_design.ghpython import get_leaf_subclasses
from timber_design.ghpython import item_input_valid_cpython
from timber_design.ghpython import manage_cpython_dynamic_params
from timber_design.ghpython import rename_cpython_gh_output
from timber_design.ghpython import warning
from timber_design.ghpython import message


class CategoryPlateJointRule(Grasshopper.Kernel.GH_ScriptInstance):
    def __init__(self):
        super(CategoryPlateJointRule, self).__init__()

        self.classes = {}
        for cls in get_leaf_subclasses(PlateJoint):
            self.classes[cls.__name__] = cls

        self.joint_type = self.classes.get(self.component.Params.Output[0].NickName, None)

    @property
    def component(self):
        return ghenv.Component  # type: ignore

    def RunScript(self, *args):
        if not self.joint_type:
            warning(self.component, "Select joint type from context menu (right click)")
            return None
        else:
            message(self.component, self.joint_type.__name__)
            cat_a, cat_b = args[:2]
            if not item_input_valid_cpython(ghenv, cat_a, self.arg_names()[0]) or not item_input_valid_cpython(ghenv, cat_b, self.arg_names()[1]):
                return

            return CategoryRule(self.joint_type, cat_a, cat_b)

    def arg_names(self):
        names = inspect.getargspec(self.joint_type.__init__)[0][1:3]
        for i in range(2):
            names[i] += "_category"
        return [name for name in names if (name != "key") and (name != "frame")] + ["max_distance"]

    def AppendAdditionalMenuItems(self, menu):
        for name in self.classes.keys():
            item = menu.Items.Add(name, None, self.on_item_click)
            if self.joint_type and name == self.joint_type.__name__:
                item.Checked = True
        menu.Items.Add(ToolStripSeparator())

    def on_item_click(self, sender, event_info):
        self.joint_type = self.classes[str(sender)]
        rename_cpython_gh_output(self.joint_type.__name__, 0, ghenv)
        manage_cpython_dynamic_params(self.arg_names(), ghenv, rename_count=2, permanent_param_count=0)
        ghenv.Component.ExpireSolution(True)



# r: timber_design>=0.1.0
# venv: td_migration
# flake8: noqa

"""Generates a feature from BTLx parameters."""

import inspect

import Grasshopper
import System
from compas.scene import Scene

from compas_timber.fabrication import BTLxProcessing
from compas_timber.fabrication import MachiningLimits
from timber_design.ghpython.ghcomponent_helpers import get_leaf_subclasses
from timber_design.ghpython.ghcomponent_helpers import manage_cpython_dynamic_params
from timber_design.ghpython.ghcomponent_helpers import rename_cpython_gh_output


class BTLxFromParams(Grasshopper.Kernel.GH_ScriptInstance):
    def __init__(self):
        super(BTLxFromParams, self).__init__()
        self.classes = {}
        for cls in get_leaf_subclasses(BTLxProcessing):
            if cls.PROCESSING_NAME == "FreeContour":
                continue
            self.classes[cls.__name__] = cls
        if ghenv.Component.Params.Output[0].NickName == "Processing":
            self.processing_type = None
        else:
            self.processing_type = self.classes.get(ghenv.Component.Params.Output[0].NickName, None)

    def RunScript(self, element: System.Collections.Generic.List[object], ref_side: System.Collections.Generic.List[int], *args):
        if not self.processing_type:
            ghenv.Component.Message = "Select Process type from context menu (right click)"
            ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning, "Select Process type from context menu (right click)")
            return None
        else:
            if not ref_side:
                ref_side = 0
            ghenv.Component.Message = self.processing_type.__name__
            kwargs = {}
            kwargs["ref_side_index"] = ref_side
            for arg, val in zip(self.arg_names(), args):
                if val is not None:
                    kwargs[arg] = val

            if "machining_limits" in self.arg_names() and kwargs.get("machining_limits") is not None:
                kwargs["machining_limits"] = kwargs["machining_limits"].limits
            processing = self.processing_type(**kwargs)

            scene = Scene()
            deferred_processings = []
            if processing and element:
                for e in element:
                    e_copy = e.copy()
                    e_copy.add_features(processing)
                    scene.add(e_copy.geometry)
                    deferred_processings.append(DeferredBTLxProcessing(processing, e))

            return deferred_processings, scene.draw()

    def arg_names(self):
        return inspect.getargspec(self.processing_type.__init__)[0][1:]

    def AppendAdditionalMenuItems(self, menu):
        for name in self.classes.keys():
            item = menu.Items.Add(name, None, self.on_item_click)
            if self.processing_type and name == self.processing_type.__name__:
                item.Checked = True

    def on_item_click(self, sender, event_info):
        self.processing_type = self.classes[str(sender)]
        rename_cpython_gh_output(self.processing_type.__name__, 0, ghenv)
        manage_cpython_dynamic_params(self.arg_names(), ghenv, rename_count=0, permanent_param_count=2)
        ghenv.Component.ExpireSolution(True)


class DeferredBTLxProcessing(object):
    def __init__(self, processing, elements):
        self.processing = processing
        self.elements = elements if isinstance(elements, list) else [elements]

    def __str__(self):
        return "DeferredBTLxProcessing({})".format(self.processing.__class__.__name__)

    def __repr__(self):
        return "DeferredBTLxProcessing(processing={!r})".format(self.processing.__class__.__name__)

    def ToString(self):
        return self.__str__()

    def feature_from_element(self, element):
        return self.processing



from .ghcomponent_helpers import add_gh_param
from .ghcomponent_helpers import clear_gh_params
from .ghcomponent_helpers import manage_dynamic_params
from .ghcomponent_helpers import list_input_valid
from .ghcomponent_helpers import item_input_valid
from .ghcomponent_helpers import get_leaf_subclasses
from .ghcomponent_helpers import get_createable_joints
from .ghcomponent_helpers import rename_gh_input
from .ghcomponent_helpers import rename_gh_output
from .ghcomponent_helpers import item_input_valid_cpython
from .ghcomponent_helpers import add_cpython_gh_param
from .ghcomponent_helpers import clear_cpython_gh_params
from .ghcomponent_helpers import manage_cpython_dynamic_params
from .ghcomponent_helpers import rename_cpython_gh_input
from .ghcomponent_helpers import rename_cpython_gh_output
from .ghcomponent_helpers import warning
from .ghcomponent_helpers import error
from .ghcomponent_helpers import remark
from .ghcomponent_helpers import message


__all__ = [
    "clear_gh_params",
    "add_gh_param",
    "manage_dynamic_params",
    "list_input_valid",
    "item_input_valid",
    "get_leaf_subclasses",
    "get_createable_joints",
    "rename_gh_input",
    "rename_gh_output",
    "item_input_valid_cpython",
    "clear_cpython_gh_params",
    "add_cpython_gh_param",
    "manage_cpython_dynamic_params",
    "rename_cpython_gh_input",
    "rename_cpython_gh_output",
    "warning",
    "error",
    "remark",
    "message",
]



# r: timber_design>=0.1.0
# venv: td_migration
# flake8: noqa

"""Creates a BTLx MachiningLimits object from boolean inputs for each face limit."""

import Grasshopper
import Rhino

from compas_timber.fabrication import MachiningLimits


class BTLxMachiningLimits(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(self, face_limited_start: bool, face_limited_end: bool, face_limited_front: bool, face_limited_back: bool, face_limited_top: bool, face_limited_bottom: bool):
        # provide default value
        inputs = [
            face_limited_start,
            face_limited_end,
            face_limited_front,
            face_limited_back,
            face_limited_top,
            face_limited_bottom,
        ]

        for i in range(len(inputs)):
            if inputs[i] is None:
                inputs[i] = True

        (
            face_limited_start,
            face_limited_end,
            face_limited_front,
            face_limited_back,
            face_limited_top,
            face_limited_bottom,
        ) = inputs

        # get MachiningLimits object
        machining_limits = MachiningLimits()
        machining_limits.face_limited_start = face_limited_start
        machining_limits.face_limited_end = face_limited_end
        machining_limits.face_limited_front = face_limited_front
        machining_limits.face_limited_back = face_limited_back
        machining_limits.face_limited_top = face_limited_top
        machining_limits.face_limited_bottom = face_limited_bottom
        return machining_limits



# r: timber_design>=0.1.0
# venv: td_migration
"""Defines which Joint type will be applied in the Automatic Joints component for connecting Beams with the given Category attributes. This overrides Topological Joint rules and is overriden by Direct joint rules"""

# flake8: noqa
import inspect
from collections import OrderedDict

import Grasshopper
from System.Windows.Forms import ToolStripMenuItem
from System.Windows.Forms import ToolStripSeparator


from timber_design.workflow import CategoryRule
from timber_design.ghpython import get_createable_joints
from timber_design.ghpython import item_input_valid_cpython
from timber_design.ghpython import manage_cpython_dynamic_params
from timber_design.ghpython import rename_cpython_gh_output
from timber_design.ghpython import warning
from timber_design.ghpython import message


class CategoryJointRule(Grasshopper.Kernel.GH_ScriptInstance):
    def __init__(self):
        super(CategoryJointRule, self).__init__()
        self.topo_bools = OrderedDict([("Unknown", False), ("I", False), ("L", False), ("T", False), ("X", False)])

        self.classes = {}
        for cls in get_createable_joints():
            self.classes[cls.__name__] = cls

        if ghenv.Component.Params.Output[0].NickName == "Rule":
            self.joint_type = None
        else:
            parsed_output = ghenv.Component.Params.Output[0].NickName.split("_")
            self.joint_type = self.classes.get(parsed_output[0])
            for key in parsed_output[1:]:
                self.topo_bools[key] = True

    @property
    def component(self):
        return ghenv.Component  # type: ignore

    def RunScript(self, *args):
        if not self.joint_type:
            warning(self.component, "Select joint type from context menu (right click)")
            return None
        else:
            message(self.component, self.joint_type.__name__)
            cat_a, cat_b = args[:2]
            kwargs = {}
            for i, val in enumerate(args[2:]):
                if val is not None:
                    kwargs[self.arg_names()[i + 2]] = val
            if not item_input_valid_cpython(ghenv, cat_a, self.arg_names()[0]) or not item_input_valid_cpython(ghenv, cat_b, self.arg_names()[1]):
                return

            topos = []
            for i, bool in enumerate(self.topo_bools.values()):
                if bool:
                    topos.append(i)

            return CategoryRule(self.joint_type, cat_a, cat_b, topos, **kwargs)

    def arg_names(self):
        names = inspect.getargspec(self.joint_type.__init__)[0][1:]
        for i in range(2):
            names[i] += "_category"
        return [name for name in names if (name != "key") and (name != "frame")] + ["max_distance"]

    def AppendAdditionalMenuItems(self, menu):
        for name in self.classes.keys():
            item = menu.Items.Add(name, None, self.on_item_click)
            if self.joint_type and name == self.joint_type.__name__:
                item.Checked = True

        menu.Items.Add(ToolStripSeparator())
        topo_menu = ToolStripMenuItem("Apply to Topology")
        menu.Items.Add(topo_menu)
        for name, bool in self.topo_bools.items():
            item = ToolStripMenuItem(name, None, self.on_topo_click)
            item.Checked = bool
            topo_menu.DropDownItems.Add(item)

    def output_name(self):
        name = self.joint_type.__name__
        for key, bool in self.topo_bools.items():
            if bool:
                name += "_{}".format(key)
        return name

    def on_topo_click(self, sender, event_info):
        self.topo_bools[str(sender)] = not self.topo_bools[str(sender)]
        rename_cpython_gh_output(self.output_name(), 0, ghenv)
        ghenv.Component.ExpireSolution(True)

    def on_item_click(self, sender, event_info):
        self.joint_type = self.classes[str(sender)]
        rename_cpython_gh_output(self.output_name(), 0, ghenv)
        manage_cpython_dynamic_params(self.arg_names(), ghenv, rename_count=2, permanent_param_count=0)
        ghenv.Component.ExpireSolution(True)



# r: timber_design>=0.1.0
# venv: td_migration
# flake8: noqa
import inspect

import Grasshopper
import System

from compas_timber.connections import Joint
from compas_timber.connections import PlateJoint
from timber_design.workflow import DirectRule
from timber_design.ghpython.ghcomponent_helpers import get_createable_joints
from timber_design.ghpython.ghcomponent_helpers import manage_cpython_dynamic_params
from timber_design.ghpython.ghcomponent_helpers import rename_cpython_gh_output
from timber_design.ghpython.ghcomponent_helpers import list_input_valid_cpython


class JointRuleFromList(Grasshopper.Kernel.GH_ScriptInstance):
    def __init__(self):
        super(JointRuleFromList, self).__init__()
        self.classes = {}
        for cls in get_createable_joints():
            if not issubclass(cls, PlateJoint):
                self.classes[cls.__name__] = cls

        self.joint_type = self.classes.get(ghenv.Component.Params.Output[0].NickName, None)

    def RunScript(self, elements: System.Collections.Generic.List[object], *args):
        if not self.joint_type:
            ghenv.Component.Message = "Select joint type from context menu (right click)"
            ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning, "Select joint type from context menu (right click)")
            return None
        else:
            ghenv.Component.Message = self.joint_type.__name__
            if not list_input_valid_cpython(ghenv, elements, self.arg_names[0]):
                return
            if not self.joint_type.element_count_complies(elements):
                ghenv.Component.AddRuntimeMessage(
                    Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning,
                    "{} requires at least {} and at most {} elements.".format(self.joint_type.__name__, self.joint_type.MIN_ELEMENT_COUNT, self.joint_type.MAX_ELEMENT_COUNT),
                )
                return
            kwargs = {}
            for i, val in enumerate(args[1:]):
                if val is not None:
                    kwargs[self.arg_names[i]] = val

            return DirectRule(self.joint_type, [e for e in elements], **kwargs)

    @property
    def arg_start_index(self):
        if self.joint_type.MAX_ELEMENT_COUNT is None:
            return 2
        elif self.joint_type.MAX_ELEMENT_COUNT == self.joint_type.MIN_ELEMENT_COUNT:
            return self.joint_type.MAX_ELEMENT_COUNT + 1
        else:
            raise Error("I don't know how to handle this joint type")

    @property
    def arg_names(self):
        return inspect.getargspec(self.joint_type.__init__)[0][self.arg_start_index :] + ["max_distance"]

    def AppendAdditionalMenuItems(self, menu):
        for name in self.classes.keys():
            item = menu.Items.Add(name, None, self.on_item_click)
            if self.joint_type and name == self.joint_type.__name__:
                item.Checked = True

    def on_item_click(self, sender, event_info):
        self.joint_type = self.classes[str(sender)]
        rename_cpython_gh_output(self.joint_type.__name__, 0, ghenv)
        manage_cpython_dynamic_params(self.arg_names, ghenv, permanent_param_count=1)
        ghenv.Component.ExpireSolution(True)



# r: timber_design>=0.1.0
# venv: td_migration
"""Creates a Beam from a LineCurve."""

# flake8: noqa
import Grasshopper
import Rhino
import rhinoscriptsyntax as rs
import System
from compas.scene import Scene
from compas_rhino.conversions import polyline_to_compas

from compas_timber.elements import Plate as CTPlate
from timber_design.ghpython.ghcomponent_helpers import item_input_valid_cpython


class PlateFromTopBottom(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(self, top, bottom, openings: System.Collections.Generic.List[object], category: str, updateRefObj: bool):
        # minimum inputs required

        if not item_input_valid_cpython(ghenv, top, "top") or not item_input_valid_cpython(ghenv, bottom, "bottom"):
            return

        scene = Scene()

        t_guid, t_geometry = self._get_guid_and_geometry(top)
        b_guid, b_geometry = self._get_guid_and_geometry(bottom)
        t_rhino_polyline = rs.coercecurve(t_geometry)
        b_rhino_polyline = rs.coercecurve(b_geometry)
        top_line = polyline_to_compas(t_rhino_polyline.ToPolyline())
        bottom_line = polyline_to_compas(b_rhino_polyline.ToPolyline())

        o = []
        if openings:
            for o_outline in openings:
                if o_outline:
                    o_guid, o_geometry = self._get_guid_and_geometry(o_outline)
                    o_rhino_polyline = rs.coercecurve(o_geometry)
                    o.append(polyline_to_compas(o_rhino_polyline.ToPolyline()))

        plate = CTPlate(top_line, bottom_line, openings=o)
        plate.attributes["rhino_guid_a"] = str(t_guid) if t_guid else None
        plate.attributes["rhino_guid_b"] = str(b_guid) if b_guid else None
        plate.attributes["category"] = category

        scene.add(plate.shape)

        geo = scene.draw()

        return plate, geo

    def _get_guid_and_geometry(self, line):  # TODO: move to ghpython_helpers
        # internalized curves and GH geometry will not have persistent GUIDs, referenced Rhino objects will
        # type hint on the input has to be 'ghdoc' for this to work
        guid = None
        geometry = line
        rhino_obj = Rhino.RhinoDoc.ActiveDoc.Objects.FindId(line)

        if rhino_obj:
            guid = line
            geometry = rhino_obj.Geometry
        return guid, geometry



# r: timber_design>=0.1.0
# venv: td_migration
"""Creates a Beam from a LineCurve."""

# flake8: noqa
import Grasshopper
import Rhino
import rhinoscriptsyntax as rs
import System
from compas.scene import Scene
from compas_rhino.conversions import polyline_to_compas
from compas_rhino.conversions import vector_to_compas

from compas_timber.elements import Plate as CTPlate
from timber_design.ghpython.ghcomponent_helpers import item_input_valid_cpython


class Plate(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(self, outline, thickness: float, vector: Rhino.Geometry.Vector3d, openings: System.Collections.Generic.List[object], category: str, updateRefObj: bool):
        # minimum inputs required

        if not item_input_valid_cpython(ghenv, outline, "Outline") or not item_input_valid_cpython(ghenv, thickness, "Thickness"):
            return
        scene = Scene()
        guid, geometry = self._get_guid_and_geometry(outline)
        rhino_polyline = rs.coercecurve(geometry)
        line = polyline_to_compas(rhino_polyline.ToPolyline())
        v = vector_to_compas(vector) if vector else None
        o = []
        if openings:
            for o_outline in openings:
                o_guid, o_geometry = self._get_guid_and_geometry(o_outline)
                o_rhino_polyline = rs.coercecurve(o_geometry)
                o.append(polyline_to_compas(o_rhino_polyline.ToPolyline()))
        plate = CTPlate.from_outline_thickness(line, thickness, vector=v, openings=o)
        plate.attributes["rhino_guid"] = str(guid) if guid else None
        plate.attributes["category"] = category

        scene.add(plate.shape)
        geo = scene.draw()
        return plate, geo

    def _get_guid_and_geometry(self, line):  # TODO: move to ghpython_helpers
        # internalized curves and GH geometry will not have persistent GUIDs, referenced Rhino objects will
        # type hint on the input has to be 'ghdoc' for this to work
        guid = None
        geometry = line
        rhino_obj = Rhino.RhinoDoc.ActiveDoc.Objects.FindId(line)

        if rhino_obj:
            guid = line
            geometry = rhino_obj.Geometry
        return guid, geometry



# r: timber_design>=0.1.0
# venv: td_migration
"""Creates a Beam from a LineCurve."""

# flake8: noqa
import Grasshopper
import Rhino
import rhinoscriptsyntax as rs
import System
from compas.scene import Scene
from compas_rhino.conversions import line_to_compas
from compas_rhino.conversions import vector_to_compas

from compas_timber.elements import Beam as CTBeam
from timber_design.ghpython.ghcomponent_helpers import list_input_valid_cpython


class Beam_fromCurve(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(
        self,
        centerline: System.Collections.Generic.List[object],
        z_vector: System.Collections.Generic.List[Rhino.Geometry.Vector3d],
        width: System.Collections.Generic.List[float],
        height: System.Collections.Generic.List[float],
        category: System.Collections.Generic.List[str],
        updateRefObj: bool,
    ):
        # minimum inputs required
        if not list_input_valid_cpython(ghenv, centerline, "Centerline"):
            return
        if not width:
            length = self._get_centerline_length(centerline)
            width = [length / 20]
        if not height:
            length = self._get_centerline_length(centerline)
            height = [length / 10]

        # reformat unset parameters for consistency
        if not z_vector:
            z_vector = [None]
        if not category:
            category = [None]

        beams = []
        blanks = []
        scene = Scene()

        if centerline and height and width:
            # check list lengths for consistency
            N = len(centerline)
            if len(z_vector) not in (0, 1, N):
                ghenv.Component.AddRuntimeMessage(
                    Grasshopper.Kernel.GH_RuntimeMessageLevel.Error, " In 'ZVector' I need either none, one or the same number of inputs as the Crv parameter."
                )
            if len(width) not in (1, N):
                ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Error, " In 'W' I need either one or the same number of inputs as the Crv parameter.")
            if len(height) not in (1, N):
                ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Error, " In 'H' I need either one or the same number of inputs as the Crv parameter.")
            if len(category) not in (0, 1, N):
                ghenv.Component.AddRuntimeMessage(
                    Grasshopper.Kernel.GH_RuntimeMessageLevel.Error, " In 'Category' I need either none, one or the same number of inputs as the Crv parameter."
                )

            # duplicate data if None or single value
            if len(z_vector) != N:
                z_vector = [z_vector[0] for _ in range(N)]
            if len(width) != N:
                width = [width[0] for _ in range(N)]
            if len(height) != N:
                height = [height[0] for _ in range(N)]
            if len(category) != N:
                category = [category[0] for _ in range(N)]

            for line, z, w, h, c in zip(centerline, z_vector, width, height, category):
                guid, geometry = self._get_guid_and_geometry(line)
                rhino_line = rs.coerceline(geometry)
                line = line_to_compas(rhino_line)

                z = vector_to_compas(z) if z else None
                beam = CTBeam.from_centerline(centerline=line, width=w, height=h, z_vector=z)
                beam.attributes["rhino_guid"] = str(guid) if guid else None
                beam.attributes["category"] = c

                beams.append(beam)
                scene.add(beam.blank)

        blanks = scene.draw()

        return beams, blanks

    def _get_guid_and_geometry(self, line):
        # internalized curves and GH geometry will not have persistent GUIDs, referenced Rhino objects will
        # type hint on the input has to be 'ghdoc' for this to work
        guid = None
        geometry = line
        rhino_obj = Rhino.RhinoDoc.ActiveDoc.Objects.FindId(line)
        if rhino_obj:
            guid = line
            geometry = rhino_obj.Geometry
        return guid, geometry

    def _get_centerline_length(self, centerline):
        centerline_length = []
        for i in centerline:
            centerline_length.append(rs.CurveLength(i))
        if centerline_length:
            length_average = sum(centerline_length) / len(centerline_length)
        else:
            length_average = 1
        return length_average



# r: timber_design>=0.1.0
# venv: td_migration
"""Creates a data-tree structure of Beams from edges of a Mesh."""

# flake8: noqa
import Grasshopper
import Rhino
import rhinoscriptsyntax as rs
from compas.geometry import Line
from compas.geometry import Vector
from compas.scene import Scene
from Grasshopper import DataTree
from Grasshopper.Kernel.Data import GH_Path

from compas_timber.elements import Beam
from timber_design.ghpython.ghcomponent_helpers import item_input_valid_cpython


class BeamTreeFromMesh(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(self, mesh: Rhino.Geometry.Mesh, width: float, height: float, category: str):
        # minimum inputs required
        if not item_input_valid_cpython(ghenv, mesh, "mesh"):
            return
        if not item_input_valid_cpython(ghenv, width, "width"):
            return
        if not item_input_valid_cpython(ghenv, height, "height"):
            return
        edge_index_pairs = []

        points = rs.MeshVertices(mesh)
        faces = rs.MeshFaceVertices(mesh)
        normals = rs.MeshFaceNormals(mesh)

        points, pt_map = set_map(points)
        faces = [[pt_map[i] for i in face] for face in faces]

        for face in faces:  # get all the edges of the mesh
            for ind in range(len(face) - 1):
                edge_set = set([face[ind - 1], face[ind]])
                if edge_set not in edge_index_pairs:
                    edge_index_pairs.append(edge_set)

        beam_Zs = []
        for pair in edge_index_pairs:  # get the z direction of the beam using the average of the adjacent face normals
            normal = Vector(0, 0, 0)
            count = 0
            for i, face in enumerate(faces):
                face = set(face)
                if pair.issubset(face):
                    count += 1
                    normal += normals[i]
            beam_Zs.append(normal * (1 / count))

        scene = Scene()
        beam_list = []
        for edge, z in zip(edge_index_pairs, beam_Zs):  # create the beams from the edges
            edge = Line(points[list(edge)[0]], points[list(edge)[1]])
            beam = Beam.from_centerline(edge, width, height, z)
            if category:
                beam.attributes["category"] = category
            beam_list.append(beam)
            scene.add(beam.blank)

        beam_tree = []
        for i in range(len(points)):  # map the beams to the vertices. Each vertex will have a list of beams that are connected to it
            joint_beams = []
            for j, pair in enumerate(edge_index_pairs):
                if i in pair:
                    joint_beams.append(beam_list[j])
            beam_tree.append(joint_beams)

        dt = DataTree[object]()
        for i, beam_branch in enumerate(beam_tree):
            dt.AddRange(beam_branch, GH_Path(i))

        blanks = scene.draw()
        return beam_list, dt, blanks


def set_map(points):
    pts_out = []
    inds_out = []
    for i, pt in enumerate(points):
        if pt not in pts_out:
            pts_out.append(pt)
        inds_out.append(pts_out.index(pt))
    return pts_out, inds_out



# r: timber_design>=0.1.0
# venv: td_migration
# flake8: noqa
import Grasshopper

from compas.scene import SceneObject
from timber_design.ghpython.ghcomponent_helpers import item_input_valid_cpython


class ShowJoiningErrors(Grasshopper.Kernel.GH_ScriptInstance):
    def RunScript(self, debug_info, i: int):
        if not item_input_valid_cpython(ghenv, i, "i"):
            i = 0
        if not item_input_valid_cpython(ghenv, debug_info, "DebugInfo"):
            return
        if not debug_info.joint_errors:
            ghenv.Component.AddRuntimeMessage(Grasshopper.Kernel.GH_RuntimeMessageLevel.Warning, "No joining errors found in input parameter 'DebugInfo'")
            return
        joining_errors = debug_info.joint_errors
        i = int(i) % len(joining_errors)
        error = joining_errors[i]

        geometries = [beam.blank for beam in error.beams]
        geometries.extend(error.debug_geometries)
        geo_objs = [SceneObject(item=geo) for geo in geometries]
        output = []
        for obj in geo_objs:
            output.extend(obj.draw())

        return error.debug_info, output
